/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "unnamed_internal.h"

static int create_list(state_t *state, const char *color, pattern_list_t *list)
{
  buffer_t buffer;
  char *pattern;
  int i, count, length, total;

double start = microtime();

  buffer_init(&buffer);
  memset(list, 0, sizeof(*list));

  if(buffer_set(&buffer, "%slist.pattern", color) != 0) return -1;
  total = count = config_maxn(state, (char*)buffer.value);
  for(i = 0; i < count; i++)
  {
    if(unlikely(buffer_set(&buffer, "%slist.pattern.%d", color, i + 1) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
      buffer_free(&buffer);
      return -1;
    }
    if(unlikely((pattern = config_get_lower(state, (char*)buffer.value)) == NULL))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to get %s", (char*)buffer.value);
      buffer_free(&buffer);
      return -1;
    }
    if(match_add(state, list, pattern) != 0)
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to add pattern %s", pattern);
      free(pattern);
      buffer_free(&buffer);
      return -1;
    }
    free(pattern);
  }

  if(buffer_set(&buffer, "%slist.domain", color) != 0)
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
    buffer_free(&buffer);
    return -1;
  }
  count = config_maxn(state, (char*)buffer.value);
  total += count;
  for(i = 0; i < count; i++)
  {
    if(unlikely(buffer_set(&buffer, "%slist.domain.%d", color, i + 1) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
      buffer_free(&buffer);
      return -1;
    }
    if(unlikely((pattern = config_get_lower(state, (char*)buffer.value)) == NULL))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to get %s", (char*)buffer.value);
      buffer_free(&buffer);
      return -1;
    }

    if(char2addr(pattern, 0, NULL) == 0)
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s is an ip address", pattern);
      free(pattern);
    }
    else if(unlikely(match_pattern_list(list, pattern) != 0))
    {
      INFO_IF(state->trace) (void)lprintf(TRACE, "%s already in list", pattern);
      free(pattern);
    }
    else
    {
      if(match_add(state, list, pattern) != 0)
      {
        ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to add pattern %s", pattern);
        free(pattern);
        buffer_free(&buffer);
        return -1;
      }
      length = (int)strlen(pattern);
      if(length + 3 <= 32)
      {
        char pattern2[32];
        pattern2[0] = '*';
        pattern2[1] = '.';
        memcpy(pattern2 + 2, pattern, (size_t)(length + 1));
        free(pattern);
        if(match_add(state, list, pattern2) != 0)
        {
          ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to add pattern %s", pattern2);
          buffer_free(&buffer);
          return -1;
        }
      }
      else
      {
        char *pattern2;
        if((pattern2 = (char*)malloc((size_t)(length + 3))) == NULL)
        {
          ERROR_IF(state->trace) (void)lprintf(TRACE, "malloc failed");
          free(pattern);
          buffer_free(&buffer);
          return -1;
        }
        pattern2[0] = '*';
        pattern2[1] = '.';
        memcpy(pattern2 + 2, pattern, (size_t)(length + 1));
        free(pattern);
        if(match_add(state, list, pattern2) != 0)
        {
          ERROR_IF(state->trace) (void)lprintf(TRACE, "failed to add pattern %s", pattern2);
          free(pattern2);
          buffer_free(&buffer);
          return -1;
        }
        free(pattern2);
      }
    }
  }
  buffer_free(&buffer);
  DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s took %f for %d patterns", color, microtime() - start, total);
  return 0;
}

int black_white_init(state_t *state)
{
  buffer_t buffer;
  int count;

  if(unlikely(create_list(state, "white", &state->white_list) != 0)) return -1;
  if(unlikely(create_list(state, "black", &state->black_list) != 0))
  {
    match_free(&state->white_list);
    return -1;
  }

  buffer_init(&buffer);
  count = config_maxn(state, "blacklist.range");
  if(unlikely((state->blacklist_range = (ip_range_t*)calloc((size_t)count, sizeof(state->blacklist_range[0]))) == NULL))
  {
    ERROR_IF(state->trace) (void)lprintf(TRACE, "calloc failed");
    match_free(&state->white_list);
    match_free(&state->black_list);
    buffer_free(&buffer);
    return -1;
  }
  for(state->blacklist_range_count = 0; state->blacklist_range_count < count; state->blacklist_range_count++)
  {
    if(unlikely(buffer_set(&buffer, "blacklist.range.%d", state->blacklist_range_count + 1) != 0))
    {
      ERROR_IF(state->trace) (void)lprintf(TRACE, "buffer_set failed");
      buffer_free(&buffer);
      black_white_term(state);
      return -1;
    }
    if(unlikely(config_get_ip_range(state, (char*)buffer.value, &state->blacklist_range[state->blacklist_range_count]) != 0))
    {
      buffer_free(&buffer);
      black_white_term(state);
      return -1;
    }
  }

  buffer_free(&buffer);
  config_free_node(state, "blacklist");
  config_free_node(state, "whitelist");
  return 0;
}

void black_white_term(state_t *state)
{
  match_free(&state->white_list);
  match_free(&state->black_list);
  if(likely(state->blacklist_range != NULL))
  {
    free(state->blacklist_range);
    state->blacklist_range_count = 0;
    state->blacklist_range = NULL;
  }
}

__attribute__((pure)) int is_ip_black(state_t *state, const address_t *a)
{
  int i;
  for(i = 0; i < state->blacklist_range_count; i++)
  {
    if(is_in_range(a, &state->blacklist_range[i]) != 0) return 1;
  }
  return 0;
}

int is_name_white(state_t *state, const buffer_t *name)
{
  int ret = match_pattern_list(&state->white_list, (char*)name->value);
  if(ret != 0) DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s is white-listed", (char*)name->value);
  return ret;
}

int is_name_black(state_t *state, const buffer_t *name)
{
  int ret;
  if(is_name_white(state, name) != 0) return 0;
  ret = match_pattern_list(&state->black_list, (char*)name->value);
  if(ret != 0) DEBUG_IF(state->trace) (void)lprintf(TRACE, "%s is black-listed", (char*)name->value);
  return ret;
}
