/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdarg.h>
#include "unnamed_internal.h"

#define BUFFER_PAD  32U

void buffer_free(buffer_t *b)
{
  if(unlikely(b == NULL)) return;
  if(unlikely(b->value != b->small_value)) free(b->value);
  buffer_init(b);
}

/* does not change length, just allocated space */
int buffer_resize(buffer_t *b, uint32_t length)
{
  uint32_t size = (length + BUFFER_PAD) & ~(BUFFER_PAD-1);
  if(unlikely(b == NULL)) return -1;
  if(unlikely(b->value == NULL)) buffer_init(b);
  if(b->allocated >= size) return 0;
  if(b->length > length) b->length = length;
  if(unlikely(length > 65536)) return -1;
  if(size < (uint32_t)sizeof(b->small_value))
  {
    if(b->value != b->small_value)
    {
      memcpy(b->small_value, b->value, (size_t)b->length);
      free(b->value);
      b->value = b->small_value;
    }
    b->allocated = (uint32_t)sizeof(b->small_value);
  }
  else if(b->value == b->small_value)
  {
    if((b->value = (uint8_t*)malloc((size_t)size)) == NULL)
    {
      b->value = b->small_value;
    }
    else
    {
      b->allocated = size;
      memcpy(b->value, b->small_value, (size_t)b->length);
    }
  }
  else if(b->allocated < size)
  {
    uint8_t *newp = b->value;
    b->value = (uint8_t*)realloc(b->value, (size_t)size);
    if(likely(b->value != NULL))
    {
      b->allocated = size;
    }
    else
    {
      b->value = newp;
    }
  }
  return (size <= b->allocated) ? 0 : -1;
}

int buffer_get_raw(buffer_t *b, uint8_t *dst, uint32_t length)
{
  uint32_t available;
  if(unlikely(b == NULL))
  {
    if(likely(dst != NULL)) memset(dst, 0, (size_t)length);
    return -1;
  }
  if(unlikely(b->value == NULL)) buffer_init(b);
  available = b->allocated - b->offset;
  if(available > length) available = length;
  if(likely(dst != NULL))
  {
    memcpy(dst, b->value + b->offset, (size_t)available);
    memset(dst + available, 0, (size_t)(length - available));
  }
  b->offset += available;
  return (available < length) ? -1 : 0;
}

int buffer_put_raw(buffer_t *b, const uint8_t *src, uint32_t length)
{
  uint32_t len;
  if(unlikely(b == NULL)) return -1;
  len = (buffer_resize(b, b->length + length) == 0) ? length : b->allocated - b->length;
  if(unlikely(src == NULL))
    memset(b->value + b->length, 0, (size_t)len);
  else
    memcpy(b->value + b->length, src, (size_t)len);
  b->length += len;
  b->value[b->length] = 0;
  return (len == length) ? 0 : -1;
}

int buffer_copy(buffer_t *dst, const buffer_t *src)
{
  if(unlikely(dst == NULL)) return -1;
  if(unlikely(src == NULL))
  {
    buffer_free(dst);
    return -1;
  }
  dst->offset = src->offset;
  dst->length = (buffer_resize(dst, src->length) == 0) ? src->length : dst->allocated - 1;
  memcpy(dst->value, src->value, (size_t)dst->length);
  dst->value[dst->length] = 0;
  return (dst->length == src->length) ? 0 : -1;
}

int buffer_get32(buffer_t *b, uint32_t *n)
{
  int ret;
  if(unlikely(n == NULL)) return -1;
  *n = 0;
  if(unlikely(b == NULL)) return -1;
  ret = buffer_get_raw(b, (uint8_t*)n, (uint32_t)sizeof(*n));
  *n = ntohl(*n);
  return ret;
}

int buffer_get16(buffer_t *b, uint16_t *n)
{
  int ret;
  if(unlikely(n == NULL)) return -1;
  *n = 0;
  ret = buffer_get_raw(b, (uint8_t*)n, (uint32_t)sizeof(*n));
  *n = (uint16_t)ntohs(*n);
  return ret;
}

int buffer_put32(buffer_t *b, uint32_t value)
{
  value = htonl(value);
  return buffer_put_raw(b, (uint8_t*)&value, (uint32_t)sizeof(value));
}

int buffer_put16(buffer_t *b, uint16_t value)
{
  value = (uint16_t)htons(value);
  return buffer_put_raw(b, (uint8_t*)&value, (uint32_t)sizeof(value));
}

int buffer_get_addr(buffer_t *b, address_t *a)
{
  memset(a, 0, sizeof(*a));
  a->a.sockaddr_in.sin_family = AF_INET;
  return buffer_get_raw(b, (uint8_t*)&a->a.sockaddr_in.sin_addr.s_addr, 4);
}

int buffer_put_addr(buffer_t *b, const address_t *a)
{
  if(a->a.sockaddr.sa_family != AF_INET) return -1;
  return buffer_put_raw(b, (const uint8_t*)&a->a.sockaddr_in.sin_addr.s_addr, 4);
}

int buffer_get_addr6(buffer_t *b, address_t *a)
{
  memset(a, 0, sizeof(*a));
  a->a.sockaddr_in6.sin6_family = AF_INET6;
  return buffer_get_raw(b, (uint8_t*)a->a.sockaddr_in6.sin6_addr.s6_addr, 16);
}

int buffer_put_addr6(buffer_t *b, const address_t *a)
{
  if(a->a.sockaddr.sa_family != AF_INET6) return -1;
  return buffer_put_raw(b, (const uint8_t*)a->a.sockaddr_in6.sin6_addr.s6_addr, 16);
}

int buffer_put_buffer(buffer_t *data, const buffer_t *b)
{
  uint32_t length;
  if(unlikely(data == NULL || b == NULL)) return -1;
  if(likely(buffer_resize(data, data->length + (int)sizeof(uint16_t) + b->length) == 0))
  {
    length = b->length;
  }
  else
  {
    if(data->allocated < data->length + (uint32_t)sizeof(uint16_t)) return -1;
    length = (data->allocated - data->length) - (int)sizeof(uint16_t);
  }
  if(unlikely(buffer_put16(data, (uint16_t)length) != 0)) return -1;
  memcpy(data->value + data->length, b->value, (size_t)length);
  data->length += length;
  return (length == b->length) ? 0 : -1;
}

int buffer_set(buffer_t *b, const char *fmt, ...)
{
  int length;
  va_list ap;
  if(unlikely(b == NULL)) return -1;
  buffer_free(b);
  if(unlikely(fmt == NULL)) return -1;
  for(;;)
  {
    va_start(ap, fmt);
    length = vsnprintf((char*)b->value, (size_t)b->allocated, fmt, ap);
    va_end(ap);
    if(likely(length >= 0)) break;
    /* some older versions of vsnprintf return a negative number if it doesn't fit. */
    if(unlikely(buffer_resize(b, b->allocated < 4096 ? b->allocated * 2 : b->allocated + 4096) != 0)) return -1;
  }
  length++;
  if((uint32_t)length > b->allocated)
  {
    if(unlikely(buffer_resize(b, (uint32_t)length) != 0)) return -1;
    (void)va_start(ap, fmt);
    length = vsnprintf((char*)b->value, (size_t)b->allocated, fmt, ap);
    va_end(ap);
    if(unlikely(length < 0)) return -1;
    length++;
  }
  b->length = (uint32_t)length - 1;
  return 0;
}

int buffer_append(buffer_t *b, const char *fmt, ...)
{
  int length, offset = b->length;
  va_list ap;
  if(unlikely(b == NULL || fmt == NULL)) return -1;
  for(;;)
  {
    va_start(ap, fmt);
    length = vsnprintf((char*)b->value + offset, (size_t)(b->allocated - offset), fmt, ap);
    va_end(ap);
    if(likely(length >= 0)) break;
    /* some older versions of vsnprintf return a negative number if it doesn't fit. */
    if(unlikely(buffer_resize(b, b->allocated < 4096 ? b->allocated * 2 : b->allocated + 4096) != 0)) return -1;
  }
  length++;
  if((uint32_t)length > b->allocated - offset)
  {
    if(unlikely(buffer_resize(b, (uint32_t)(length + offset)) != 0)) return -1;
    (void)va_start(ap, fmt);
    length = vsnprintf((char*)b->value + offset, (size_t)(b->allocated - offset), fmt, ap);
    va_end(ap);
    if(unlikely(length < 0)) return -1;
    length++;
  }
  b->length = (uint32_t)(offset + length - 1);
  return 0;
}

__attribute__((pure)) int buffer_cmp(const buffer_t *a, const buffer_t *b)
{
  uint32_t i, len;
  if(unlikely(a == NULL))
  {
    if(unlikely(b == NULL)) return 0;
    return -1;
  }
  if(unlikely(b == NULL)) return 1;
  len = (a->length < b->length) ? a->length : b->length;
  for(i = 0; i < len; i++)
  {
    int al = tolower((int)a->value[i]), bl = tolower((int)b->value[i]);
    if(al < bl) return -1;
    if(al > bl) return 1;
  }
  if(a->length < b->length) return -1;
  if(a->length > b->length) return 1;
  return 0;
}
