/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"
#include "unnamed.h"

static int instance = 0;

typedef struct
{
  char* lua_src;
  uint32_t mx_preference;
  uint32_t soa_refresh, soa_retry, soa_expire, soa_minimum;
  uint32_t ttl;
  int trace;
} module_data_t;

static void term(int module_index, state_t *state)
{
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  free((void*)module->lua_src);
  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  buffer_t name;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  instance++;
  if(module == NULL) return -1;
  module->trace = state->config_get_int(state, "lua.trace", TRACE_LEVEL_ERROR);
  module->ttl = (uint32_t)state->config_get_int(state, "lua.ttl", DEFAULT_TTL);
  module->mx_preference = (uint32_t)state->config_get_int(state, "lua.mx_preference", 10);
  module->soa_refresh = (uint32_t)state->config_get_int(state, "lua.soa_refresh", (int)module->ttl);
  module->soa_retry = (uint32_t)state->config_get_int(state, "lua.soa_retry", (int)module->soa_refresh);
  module->soa_expire = (uint32_t)state->config_get_int(state, "lua.soa_expire", (int)module->soa_refresh);
  module->soa_minimum = (uint32_t)state->config_get_int(state, "lua.soa_minimum", (int)module->soa_refresh);
  buffer_init(&name);
  if(state->buffer_set(&name, "lua.src.%d", instance) != 0)
  {
    state->buffer_free(&name);
    free(module);
    return -1;
  }
  if((module->lua_src = state->config_get(state, (char*)name.value)) == NULL)
  {
    state->buffer_free(&name);
    free(module);
    return -1;
  }
  state->module_data[module_index] = module;
  state->config_free_node(state, (char*)name.value);
  state->buffer_free(&name);
  return 0;
}

static int get_int(lua_State* L, const char *key, int value)
{
  lua_pushstring(L, key);
  lua_gettable(L, -2);
  if(lua_isnumber(L, -1) != 0) value = (int)lua_tointeger(L, -1);
  lua_pop(L, 1);
  return value;
}

static void get_buffer(lua_State* L, state_t *state, const char *key, buffer_t *dst, const buffer_t *value)
{
  lua_pushstring(L, key);
  lua_gettable(L, -2);
  if(lua_isstring(L, -1) != 0)
  {
    size_t len = 0;
    const char *val = lua_tolstring(L, -1, &len);
    if(state->buffer_resize(dst, (uint32_t)len) == 0)
    {
      memcpy(dst->value, val, len);
      dst->length = (uint32_t)len;
    }
  }
  else if(value != NULL)
  {
    (void)state->buffer_copy(dst, value);
  }
  lua_pop(L, 1);
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  lua_State* L;
  answer_t ret = ANSWER_UNSATISFIED;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  if(module->lua_src == NULL) return ANSWER_UNSATISFIED;
  L = lua_open();
  luaL_openlibs(L);
  if(luaL_loadfile(L, module->lua_src) != 0)
  {
    lua_close(L);
    return ANSWER_UNSATISFIED;
  }
  if(lua_pcall(L, 0, -1, 0) == 0)
  {
    int err;
    lua_getglobal(L, "answer");
    lua_pushlstring(L, (char*)name->value, name->length);
    lua_pushinteger(L, (lua_Integer)type);
    lua_pushinteger(L, (lua_Integer)dclass);
    if(lazy != 0)
    {
      lua_pushinteger(L, (lua_Integer)lazy);
      err = lua_pcall(L, 4, 1, 0);
    }
    else
    {
      err = lua_pcall(L, 3, 1, 0);
    }
    if(err != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "%s", lua_tostring(L, -1));
    }
    else if(lua_istable(L, -1) != 0)
    {
      for(lua_pushnil(L); lua_next(L, -2) != 0; lua_pop(L, 1))
      {
        if(lua_istable(L, -1) != 0)
        {
          rr_t* rr = NULL;
          buffer_t current_name;
          unsigned int current_ttl = (unsigned int)get_int(L, "ttl", (int)module->ttl);
          int current_type = get_int(L, "type", (int)type);
          int current_class = get_int(L, "class", (int)dclass);
          buffer_init(&current_name);
          get_buffer(L, state, "name", &current_name, name);

          rr = state->rr_create(&current_name, (uint16_t)current_type, (uint16_t)current_class, current_ttl + (uint32_t)time(NULL), RCODE_NO_ERROR);
          state->buffer_free(&current_name);
          if(rr != NULL)
          {
            if(current_class == CLASS_IN)
            {
              switch(current_type)
              {
                case TYPE_A:
                case TYPE_AAAA:
                {
                  lua_pushstring(L, "address");
                  lua_gettable(L, -2);
                  if(lua_isstring(L, -1) != 0)
                  {
                    const char *ip = lua_tostring(L, -1);
                    if(state->char2addr(ip, 0, &rr->p.address) == 0)
                    {
                      if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                    }
                  }
                  lua_pop(L, 1);
                } break;
                case TYPE_CNAME:
                {
                  get_buffer(L, state, "cname", &rr->p.cname.name, name);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_MX:
                {
                  get_buffer(L, state, "exchange", &rr->p.mx.exchange.name, name);
                  rr->p.mx.preference = (uint16_t)get_int(L, "preference", (int)module->mx_preference);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_NS:
                {
                  get_buffer(L, state, "ns", &rr->p.ns.name, name);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_PTR:
                {
                  get_buffer(L, state, "ptr", &rr->p.ptr.name, name);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_SOA:
                {
                  get_buffer(L, state, "mname", &rr->p.soa.mname.name, name);
                  get_buffer(L, state, "rname", &rr->p.soa.rname.name, name);
                  rr->p.soa.serial = (uint32_t)get_int(L, "serial", 0);
                  if(rr->p.soa.serial == 0) rr->p.soa.serial = state->current_serial();
                  rr->p.soa.refresh = (uint32_t)get_int(L, "refresh", (int)module->soa_refresh);
                  rr->p.soa.retry = (uint32_t)get_int(L, "retry", (int)module->soa_retry);
                  rr->p.soa.expire = (uint32_t)get_int(L, "expire", (int)module->soa_expire);
                  rr->p.soa.minimum = (uint32_t)get_int(L, "minimum", (int)module->soa_minimum);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_SPF:
                {
                  get_buffer(L, state, "spf", &rr->p.spf, NULL);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
                case TYPE_TXT:
                {
                  get_buffer(L, state, "txt", &rr->p.txt, NULL);
                  if(state->add_rr(client, rr_section, &rr) == 0) ret = ANSWER_SATISFIED;
                } break;
              }
            }
            state->free_rr(&rr);
          }
        }
      }
    }
  }
  lua_close(L);
  return ret;
}

module_t unnamed_module = { "lua", NULL, init, answer, NULL, term };
